/* ------------------------- linec.c ---------------------------------------- */

/*
 * Draw lines on SVGA. Can be cery fast when usage of low level assembly
 * support is requested. It handles VGA page crossing.
 *
 * It originated from the DJgpp package but completely re-written since, the
 * only surviving clue being the WriteMode specification in the color field.
 * Actually this is not used either.
 *
 * These functions are pure C:
 *
 * GrMoveToC (x, y)
 * GrDrawToC (x, y, color)
 * GrLineC (x0, y0, x1, y1, color)
 *
 *
 * If you only use the low-level asm version then this module is redundant.
*/

#include <stdlib.h>

#include "grh.h"

#define FAST_PLOT(va) \
	if (!_GrWriteMode)			\
		*va = (uchar)c;			\
	else if (_GrWriteMode & GrXOR)		\
		*va ^= c;			\
	else /* if (_GrWriteMode & GrXOR) */	\
		*va |= c

#define LINE_PLOT(va,vb) \
	if ((int)(MY_OFF(va)^vb) < 0) {				\
		if (MY_OFF(va) > 0xc000U)			\
			--Page, _GrSetPage (Page, Page);	\
		else if (MY_OFF(va) < 0x4000U)			\
			++Page, _GrSetPage (Page, Page);	\
	}							\
	FAST_PLOT (va)

extern uchar far * far _fastcall
_GrAddrC (x, y)
uint	x;
uint	y;
{
	ulong	xy;
	uint	page;

	page = (uint)((xy = x + _ActiveBase + y*(long)_GrSizeX) >> 16);
#if 0
	if (page > 0x0f)
		return VGA_PAGE;
#endif
	if (page != _ReadPage)
		_GrSetPage (page, page);

	return (VGA_PAGE + (uint)xy);
}

static int far
_InnerLoop (iva, dy, dx, dvx, dvy, c)
uint	iva;
int	dy, dx, dvx, dvy, c;
{
	register long	err;		/* conceptualy */
	register int	count;
	uchar		far *va = VGA_PAGE;
	uint		vb;
	int		Page = _ReadPage;

	/* dx > dy  never =!!!	*/
	/* dx > 0 		*/
	/* dy >= 0		*/

	/* set:			*/
	/* dy = (dy*0x10000)/dx	*/
	/* dx = 0x10000		*/

	MY_OFF(va) = iva;
	count = dx;

#ifdef USE_DIVTAB
	dy = dy * divtab[dx];
#else
	dy = (int)((0x00010000L*dy)/dx);
#endif

	err = 0x00008000L;
	for (; --count >= 0;) {
		vb = MY_OFF(va);
		MY_OFF(va) += dvx;
		if ((err += (ulong)(uint)dy) & 0xffff0000L) {
			err &= 0x0000ffffL;
			MY_OFF(va) += dvy;
		}
		LINE_PLOT (va,vb);
	}
	_ReadPage = Page;
	return (MY_OFF(va));
}

#if 0
static int far
_InnerLoopOld (iva, dy, dx, dvx, dvy, c)
uint	iva;
int	dy, dx, dvx, dvy, c;
{
	register int	count, err;
	uchar		far *va = VGA_PAGE;
	uint		vb;
	int		Page = _ReadPage;

	MY_OFF(va) = iva;
	err = dx / 2;
	for (count = dx; --count >= 0;) {
		vb = MY_OFF(va);
		MY_OFF(va) += dvx;
		err += dy;
		if (err > dx) {
			err -= dx;
			MY_OFF(va) += dvy;
		}
		LINE_PLOT (va,vb);
	}
	_ReadPage = Page;
	return (MY_OFF(va));
}
#endif

static int far
_SimpleLoop (vi, count, sv, c)
uint		vi;
register int	count;
register int	sv;
int		c;
{
	uchar		far *va = VGA_PAGE;
	uint		vb;
	int		Page = _ReadPage;
	

	MY_OFF(va) = vi;
	while (--count >= 0) {
		vb = MY_OFF(va);
		MY_OFF(va) += sv;
		LINE_PLOT (va,vb);
	}
	_ReadPage = Page;
	return (MY_OFF(va));
}

static uint	x1 = 0, y1 = 0, xydone = 0;
static uchar	far *va = VGA_PAGE;

extern void far
GrMoveToC (x, y)
uint	x;
uint	y;
{
#if 0
	if (x >= _GrSizeX)
		x = _GrSizeX-1;
	if (y >= _GrSizeY)
		y = _GrSizeY-1;
#endif
	x1 = x;
	y1 = y;
	xydone = 0;
}

extern void far
GrDrawToC (x2, y2, c)
uint	x2;
uint	y2;
uint	c;
{
	register int	dx, dy, svx, svy;
	int		t;

#if 0
	if (x2 >= _GrSizeX)
		x2 = _GrSizeX-1;
	if (y2 >= _GrSizeY)
		y2 = _GrSizeY-1;
#endif
	t = 0;
	if (!xydone) {
		t = 1;
		xydone = 1;
		MY_OFF(va) = (uint)_GrAddrC (x1, y1);
		FAST_PLOT(va);
	}

	svx = 1;
	if ((dx = x2 - x1) < 0) {
		dx = -dx;
		svx = -svx;
	}

	svy = _GrSizeX;
	if ((dy = y2 - y1) < 0) {
		dy = -dy;
		svy = -svy;
	}

	if (dx > dy) {
		if (0 == dy) {
			uint	vb;

			vb = MY_OFF(va) + x2 - x1;	/* end point */

			if (!((int)(MY_OFF(va)^vb) < 0 &&
					(vb > 0xc000U || vb < 0x4000U))
			&& !_GrWriteMode
			) {
				memset (VGA_PAGE + (svx>0 ? MY_OFF(va)+1 : vb),
					c, dx);
				MY_OFF(va) = vb;
			} else
				MY_OFF(va) = _SimpleLoop (MY_OFF(va), dx, svx,
						c);
		} else
			MY_OFF(va) = _InnerLoop (MY_OFF(va), dy, dx, svx, svy,
					c);
		++GrStats[dx+t];
	} else if (dx < dy) {
		if (0 == dx)
			MY_OFF(va) = _SimpleLoop (MY_OFF(va), dy, svy, c);
		else
			MY_OFF(va) = _InnerLoop (MY_OFF(va), dx, dy, svy, svx,
					c);
		++GrStats[dy+t];
	} else /* if (dx == dy) */ {
		if (dx)
			MY_OFF(va) = _SimpleLoop (MY_OFF(va), dx, svx+svy, c);
		++GrStats[dx+t];
	}

	x1 = x2;
	y1 = y2;
}


extern void far
GrLineC (xa, ya, xb, yb, c)
uint	xa;
uint	ya;
uint	xb;
uint	yb;
uint	c;
{
	if (xa != x1 || ya != y1 || !xydone)
		GrMoveToC (xa, ya);
#if 0
	if ((c&0xff00) != _GrWriteMode)
		GrSetWriteMode (c);
#endif
	GrDrawToC (xb, yb, c);
}
